#include "util.h"
#include "ENLS.h"

inline double softThreshold(double x, double lambda) {
  return(sign(x) * fmax2(fabs(x) - lambda, 0.));
}

void updateBeta(double *X, double *Y, double *beta,
		double *Xbeta, double *partialResidual,
		int n, int p, double lambda, double alpha, double *avgVar) {
  int i, j;
  const char transX = 'N';
  const double alp = 1.;
  const double bet = 0.;
  const int one = 1;
  double la = lambda * alpha;
  double loma = lambda * (1. - alpha);
  
  for (j = p; j--; ) {
    beta[j] = 0.;
    dgemv_(&transX, &n, &p, &alp, X, &n, beta, &one, &bet, Xbeta, &one);
    for (i = n; i--; )
      partialResidual[i] = Y[i] - Xbeta[i];
    beta[j] = softThreshold(innerProduct(&X[n*j], partialResidual, n)/((double)n), la) / (avgVar[j] + loma);  
  }
}

void updateBetaInactiveSet(double *X, double *Y, double *beta,
			   double *Xactive, double *betaActive, int sizeActiveSet,
			   double *Xbeta, double *partialResidual,
			   int *inactiveSet, int sizeInactiveSet,
			   int n, int p, double lambda, double alpha, double *avgVar) {
  int i, j;
  const char transX = 'N';
  const double alp = 1.;
  const double bet = 0.;
  const int one = 1;
  int jIS;
  double la = lambda * alpha;
  double loma = lambda * (1. - alpha);

  dgemv_(&transX, &n, &sizeActiveSet, &alp, Xactive, &n, betaActive, &one, &bet, Xbeta, &one);
  for (i = n; i--; )
    partialResidual[i] = Y[i] - Xbeta[i];
  
  for (j = sizeInactiveSet; j--; ) {
    jIS = inactiveSet[j];
    beta[jIS] = softThreshold(innerProduct(&X[n*jIS], partialResidual, n)/((double)n), la) / (avgVar[jIS] + loma);  
  }
}

void updateActiveAndInactiveSets(int *activeSet, int *sizeActiveSet, int *inactiveSet, int *sizeInactiveSet,
				 double *beta, int p) {
  int j;
  *sizeActiveSet = 0;
  *sizeInactiveSet = 0;
  for (j = 0; j < p; j++) {
    if (fabs(beta[j]) > 0.) {
      activeSet[*sizeActiveSet] = j;
      (*sizeActiveSet)++;
    } else {
      inactiveSet[*sizeInactiveSet] = j;
      (*sizeInactiveSet)++;
    }
  }
}

// updateXactive assumes sizeActiveSet > 0!!!
// updateXactive assumes activeSet indexes from 0.
void updateXactive(double *X, double *Xactive, int nrowsX, int *activeSet, int sizeActiveSet) {
  int j, k;
  for (j = sizeActiveSet; j--; )
    for (k = nrowsX; k--; )
      Xactive[nrowsX*j + k] = X[nrowsX*activeSet[j] + k];
}

int minimizeElasticNetSquaredLoss(double *X, double *Y, double *beta, int n, int p, 
		 double lambda, double alpha, int maxiter, double tol) {
  int i, j, iter;
  double *Xbeta = calloc(n, sizeof(double));
  double *avgVar = calloc(p, sizeof(double));
  double *avgVarActive = calloc(p, sizeof(double));
  double *betaLast = calloc(p, sizeof(double));
  double *partialResidual = calloc(n, sizeof(double));
  double *Xactive = calloc( n * (p), sizeof(double));
  double *betaActive = calloc(p, sizeof(double));
  int *activeSet = calloc(p, sizeof(int));
  int *inactiveSet = calloc(p, sizeof(int));
  int sizeActiveSet, sizeInactiveSet, inner_iter_sum=0;
  double sum;

  for (j = p; j--; )
    avgVar[j] = innerProduct(&X[n*j], &X[n*j], n)/((double)n);
  
  // Initialize active set variables
  updateActiveAndInactiveSets(activeSet, &sizeActiveSet, inactiveSet, &sizeInactiveSet, beta, p);
  
  for (;;) {
    if (sizeActiveSet > 0) {
      updateXactive(X, Xactive, n, activeSet, sizeActiveSet);
      copyBigVectorIntoSmallVector(beta, betaActive, activeSet, sizeActiveSet);
      
      for (j = 0; j < sizeActiveSet; j++)
	avgVarActive[j] = avgVar[activeSet[j]];
      
      // Run coordinate descent on active set.
      for (iter = maxiter; iter--; ) {
	//	copyArray(betaActive, betaLast, sizeActiveSet);
	updateBeta(Xactive, Y, betaActive, Xbeta, partialResidual, n, sizeActiveSet, lambda, alpha, avgVarActive);

	// Check Inner Loop Convergence
	/*	if (hasConverged(betaActive, betaLast, sizeActiveSet, tol)) {
	  //	  Rprintf("Number of inner iterations = %d.\n", maxiter - iter);
	  break;
	  }*/
      }
      copySmallVectorIntoBigVector(betaActive, beta, activeSet, sizeActiveSet);
      inner_iter_sum += fmin2(maxiter - iter, maxiter);
    }
    updateBetaInactiveSet(X, Y, beta, Xactive, betaActive, sizeActiveSet, Xbeta, partialResidual,
			  inactiveSet, sizeInactiveSet, n, p, lambda, alpha, avgVar);
    sum = 0.;
    for (j = sizeInactiveSet; j--; )
      sum += fabs(beta[inactiveSet[j]]);
    
    // Variables may drop out or come in.
    updateActiveAndInactiveSets(activeSet, &sizeActiveSet, inactiveSet, &sizeInactiveSet, beta, p);
    
    if (sum == 0.)
      break;
  }
  free(Xbeta);
  free(avgVar);
  free(avgVarActive);
  free(betaLast);
  free(partialResidual);
  free(Xactive);
  free(betaActive);
  free(activeSet);
  free(inactiveSet);
  
  return(inner_iter_sum);

}

int ENLS(double *X, double *Y, double *beta, int *n, int *p, 
		 double *lambda, double *alpha, int *maxiter, double *tol) {

  return(minimizeElasticNetSquaredLoss(X, Y, beta, *n, *p, *lambda, *alpha, *maxiter, *tol));

}
