# Description: This script generates data for demonstrating breakdown behavior of the L2E estimator
#              compared to the MLE. A total of 2n + nc points are generated.
#
# Author: Eric Chi
# Date: 6/1/2010
#
# Input:
#  n		Number of samples in one subpopulation out of two subpopulations.
#  p		Number of covariates
#  nc		Number of contaminants
#  shift	The offset of the outlier in all p signal directions.
#  ss		Signal strength is the mean for one subpopulation and the negative of the mean of the other.
#  x.sd		The standard deviation of each subpopulation.
#  beta		The coefficient vector.
#  seed		An unsigned int, random seed.
#
# Output:
#  X is a 2n+nc-by-p+1 matrix of covariates.
#  Y is a binary response vector of length 2n+nc.

genLogit = function(n=100, p=4, nc=5, shift=0.75, ss=0.25, x.sd=0.4,
	beta=matrix(c(0,1,0.5,1,2),ncol=1), seed=111908) {

	if (p != length(beta) - 1) stop("p must equal length(beta) - 1")
	set.seed(seed)
	A = matrix(ss,n,p)
	mu = rbind(-A,A)

	X = matrix(rnorm(2*n*p,0,x.sd),ncol=p) + mu
	X = apply(X,2,FUN=function(x){x-mean(x)})
	X = cbind(1,X)

	P = plogis(X %*% beta)
	Y = ifelse(runif(2*n) <= P,1,0)

	# Add junk
	if (nc > 0) {
		X = rbind(X,matrix(NA,nc,p+1))
		X[(2*n+1):(2*n+nc), 1] = 1
		X[(2*n+1):(2*n+nc), -1] = shift
		Y = rbind(Y, matrix(0,nc,1))
	}

	return(list(beta=beta,X=X,Y=Y))
}