# Simulation Experiments for 7.3.
# "Variable Selection in High Dimensions."
#
# Note that this script can take a long time to run. ~10 hours
#   on Intel(R) Xeon(R) CPU X5680  @ 3.33GHz

rm(list=ls())

pwd = getwd()
setwd("../../Rfunctions")
source("rLogistic.R")
source("generateVariableSelectionData.R")
source("calculateLambdaSequence.R")
source("hhsvm.R")
source("median.l2e.R")
source("median.devLoss.R")
setwd(pwd)
library(glmnet)

## MM algorithm
loss = function(u, t=0.5) {
	loss = ((1-t)^2 + 2*(1-t)*(t-u))*ifelse(u <= t, 1, 0)
        loss = loss + (1-u)^2 * ifelse(t < u & u <= 1,1,0)
	return(loss)
}

lp = function(u, t=0.5) {
	lp = -2*((1-t)*ifelse(u <= t, 1, 0) + (1-u)*ifelse(u > t & u <= 1, 1, 0))
	return(lp)
}

## Fit intercept
fitIntercept = function(b0, Y) {
	return(sum(loss(Y*b0)))	
}

numFP = function(true, test) {
	length(setdiff(test,true))
}

numTP = function(true, test) {
	length(true) - length(setdiff(true,test))
}

# Create simulated data scenario
delta = 1
nc = 100
#nc = 0
seed = 6
alpha = 0.6
p.true = 50; p.dummy = 500-p.true
n.true = 200
x.sd = 0.75
w = 1

L2E.nLambda = 50

### K-fold Cross-Validation
nFolds = 10
nObs = 2*n.true+nc
holdOutSets = matrix(NA, nObs/nFolds, nFolds)
set.seed(1119)
theSet = 1:nObs
for (i in 1:nFolds) {
  hos = sample(theSet, nObs/nFolds)
  theSet = setdiff(theSet, hos)
  holdOutSets[,i] = hos
}

nReps = 10
L2E.FP = matrix(0, nReps)
L2E.TP = matrix(0, nReps)
MLE.FP = matrix(0, nReps)
MLE.TP = matrix(0, nReps)
HHSVM.FP = matrix(0, nReps)
HHSVM.TP = matrix(0, nReps)

L2E_CV_HX = vector(mode="list", nReps)
MLE_CV_HX = vector(mode="list", nReps)
HHSVM_CV_HX = vector(mode="list", nReps)

L2E_lambda_HX = vector(mode="list", nReps)
MLE_lambda_HX = vector(mode="list", nReps)
HHSVM_lambda_HX = vector(mode="list", nReps)

L2E_models = vector(mode="list", nReps)
MLE_models = vector(mode = "list", nReps)
HHSVM_models = vector(mode = "list", nReps)

for (r in 1:nReps) {
	L2E_models[[r]]	= vector(mode = "list", nFolds)
	MLE_models[[r]]	= vector(mode = "list", nFolds)
	HHSVM_models[[r]] = vector(mode = "list", nFolds)
}

for (repIndex in 1:nReps) {

	###### Make Data
	data = generateVariableSelectionData(seed=sample(1:111908,1), delta = delta, nc = nc, n = n.true, p = 	p.true, x.sd = x.sd, p.dummy=p.dummy)
	X = data$X
	Y = data$Y

	n = nrow(X); p = ncol(X)
	Z = scale(X)

	####### A. L2E ####################################

	lambda = calculateLambdaSequence(Y, Z, alpha=alpha, w=1, epsilon=0.02, nLambda=L2E.nLambda)
	L2E.CV = matrix(NA, nFolds, L2E.nLambda)

	###### Compute scores
	p0 = mean(Y)
	scores = p0*(1-p0)*abs(t(Z) %*% (Y-p0))
	ix = which(scores >= quantile(scores,0.90))

	## Run the L2E fit forward
	beta0 = log(mean(Y)/(1-mean(Y)))
	beta = matrix(0,p,1)
	beta[ix] = 1
	system.time({rLogistic.fit = rLogistic(Z,Y,alpha,lambda,beta0,beta,maxVar=250, w=w)})

	## Run the L2E fit backwards
	lambda_first = rLogistic.fit$lambda[length(rLogistic.fit$lambda)]/2
	lambda_last = rLogistic.fit$lambda[1]
	lambda = 10^seq(log10(lambda_first), log10(lambda_last), length.out=L2E.nLambda)

	L2E_lambda_HX[[repIndex]] = lambda

	###### Create Test/Fit sets
	for (iSet in 1:nFolds) {
		holdOut = holdOutSets[,iSet]
		Zfit = X[-holdOut,,drop=F]
		Yfit = Y[-holdOut]

		###### Compute scores
		p0 = mean(Yfit)
		scores = p0*(1-p0)*abs(t(Zfit) %*% (Yfit-p0))
		ix = which(scores >= quantile(scores,0.90))

		##### A. Calculate L2E regularization path solutions
		beta = matrix(0, p, 1)
		beta[ix] = 1
		beta0 = log(mean(Yfit)/(1-mean(Yfit)))
		system.time({rLogistic.fit = rLogistic(Zfit,Yfit,alpha,lambda,beta0,beta,niter=10000,maxVar=p, w=w)})
		L2E_models[[repIndex]][[iSet]] = rLogistic.fit

		##### Calcualte L2E loss values for path solutions with no regularization.
		for (i in 1:rLogistic.fit$dim[2]) {
			varSel = which(sign(rLogistic.fit$beta[,i]) != 0)
			beta = matrix(1, length(varSel), 1)
			beta0 = log(mean(Yfit)/(1-mean(Yfit)))
			system.time({rLogistic.cv = rLogistic(Zfit[,varSel,drop=F],Yfit,alpha=0,lambda=0.01,beta0,beta,niter=10000,maxVar=p, w=w)})
			L2E.CV[iSet,i] = median.l2e(w, rLogistic.cv$a0, as.matrix(rLogistic.cv$beta), Z[holdOut,varSel,drop=F], Y[holdOut])
		}
		#### Calculate summaries of the L2E runs.
		print(paste("L2E: Fold ", iSet, "; completed!"))
	}

	u = apply(L2E.CV,2,median)
	md = apply(L2E.CV,2,mad)
	minIX = which(u == min(u))
	lambda_star = lambda[max(which(u <= u[minIX] + md[minIX]))]
	
	p0 = mean(Y)
	scores = p0*(1-p0)*abs(t(Z) %*% (Y-p0))
	ix = which(scores >= quantile(scores,0.90))
	beta0 = log(mean(Y)/(1-mean(Y)))
	beta = matrix(0,p,1)
	beta[ix] = 1
	
	L2E.final = rLogistic(Z,Y,alpha,lambda=lambda_star, beta0, beta, niter=10000,maxVar=p, w=w)
	varSel.final = which(sign(L2E.final$beta) != 0)
	L2E.FP[repIndex] = sum(!varSel.final %in% 1:p.true)
	L2E.TP[repIndex] = sum(varSel.final %in% 1:p.true)
	L2E_CV_HX[[repIndex]] = L2E.CV
	
	####### B. MLE ####################################
	system.time({glmnet.fit = glmnet(Z,Y,family="binomial",alpha=alpha,standardize=T)})
	lambda.mle = glmnet.fit$lambda
	MLE.CV = matrix(NA, nFolds, length(lambda.mle))
	MLE_lambda_HX[[repIndex]] = lambda.mle

	###### Create Test/Fit sets
	for (iSet in 1:nFolds) {
		holdOut = holdOutSets[,iSet]
		Zfit = X[-holdOut,,drop=F]
		Yfit = Y[-holdOut]
	
		system.time({glmnet.fit = glmnet(Zfit,Yfit,family="binomial",alpha=alpha,lambda=lambda.mle, standardize=T)})
		MLE_models[[repIndex]][[iSet]] = glmnet.fit

		##### Calcualte MLE loss values for path solutions with no regularization.
		for (i in 1:length(glmnet.fit$lambda)) {
			varSel = which(sign(glmnet.fit$beta[,i]) != 0)
			if (length(varSel) == 0) {
				MLE.CV[iSet,i] = median(-Y[holdOut]*mean(Yfit) + log(1 + exp(mean(Yfit))))
			} else {
				system.time({glmnet.cv = glmnet(Zfit[,varSel,drop=F],Yfit,family="binomial",alpha=0,
					lambda=10*min(glmnet.fit$lambda),standardize=T)})
				MLE.CV[iSet,i] = median.devLoss(glmnet.cv$a0, as.matrix(glmnet.cv$beta), Z[holdOut,varSel,drop=F], Y[holdOut])
			}
		}
		#### Calculate summaries of the L2E runs.
		print(paste("MLE: Fold ", iSet, "; completed!"))
	}
	u.mle = apply(MLE.CV,2,median,na.rm=T)
	md.mle = apply(MLE.CV,2,mad,na.rm=T)
	minIX.mle = which(u.mle == min(u.mle))
	lambda.mle_star = lambda.mle[min(which(u.mle <= u.mle[minIX.mle] + md.mle[minIX.mle]))]
	
	MLE.final = glmnet(Z,Y,family="binomial",alpha=alpha, lambda=lambda.mle_star, standardize=T)
	varSel.MLE.final = which(sign(MLE.final$beta) != 0)
	MLE.FP[repIndex] = sum(!varSel.MLE.final %in% 1:p.true)
	MLE.TP[repIndex] = sum(varSel.MLE.final %in% 1:p.true)
	MLE_CV_HX[[repIndex]] = MLE.CV

	####### C. HHSVM ####################################
	y = 2*Y-1
	beta0 = nlminb(1, fitIntercept, Y=y)$par
	beta = matrix(0, ncol(X))
	lambdaMax = max(abs(t(Z) %*% (y*lp(y*beta0))))/(2*nrow(Z)*alpha)
	eps = 0.05
	lambdaMin = eps*lambdaMax
	nLambda = 100
	lambda = 10^(seq(log10(lambdaMax), log10(lambdaMin), length.out=nLambda))

	system.time({hhsvm.fit = hhsvm(Z,2*Y-1,alpha,lambda,beta0,beta,niter=10000,maxVar=250)})
	
	lambda.hhsvm = hhsvm.fit$lambda
	HHSVM.CV = matrix(NA, nFolds, length(lambda.hhsvm))
        HHSVM_lambda_HX[[repIndex]] = lambda.hhsvm	

	###### Create Test/Fit sets
	for (iSet in 1:nFolds) {
		holdOut = holdOutSets[,iSet]
		Zfit = X[-holdOut,,drop=F]
		Yfit = Y[-holdOut]
		
		beta = matrix(0, p, 1)
		beta0 = log(mean(Yfit)/(1-mean(Yfit)))
	
		system.time({hhsvm.fit = hhsvm(Zfit,2*Yfit-1,alpha,lambda.hhsvm,beta0,beta,niter=10000,maxVar=250)})
		HHSVM_models[[repIndex]][[iSet]] = hhsvm.fit
		##### Calcualte HHSVM loss values for path solutions with no regularization.
		for (i in 1:length(hhsvm.fit$lambda)) {
			varSel = which(sign(hhsvm.fit$beta[,i]) != 0)
			if (length(varSel) == 0) {
				HHSVM.CV[iSet,i] = median(-Y[holdOut]*mean(Yfit) + log(1 + exp(mean(Yfit))))
			} else {
				beta = matrix(1, length(varSel), 1)
				beta0 = log(mean(Yfit)/(1-mean(Yfit)))
				system.time({hhsvm.cv = hhsvm(Zfit[,varSel,drop=F],2*Yfit-1,alpha=0,
					lambda=10*min(hhsvm.fit$lambda),beta0, beta,niter=10000,maxVar=p)})
				hhsvm.LP = (2*Y[holdOut]-1)*(hhsvm.cv$a0 + Z[holdOut,varSel,drop=F] %*% as.matrix(hhsvm.cv$beta))
				
				
				HHSVM.CV[iSet,i] = median(apply(matrix(hhsvm.LP,ncol=1),2,FUN=loss))
				}
		}
		#### Calculate summaries of the HHSVM runs.
		print(paste("HHSVM: Fold ", iSet, "; completed!"))
	}
	u.hhsvm = apply(HHSVM.CV,2,median,na.rm=T)
	md.hhsvm = apply(HHSVM.CV,2,mad,na.rm=T)
	minIX.hhsvm = which(u.hhsvm == min(u.hhsvm))
	lambda.hhsvm_star = lambda.hhsvm[min(which(u.hhsvm <= u.hhsvm[minIX.hhsvm] + md.hhsvm[minIX.hhsvm]))]
	beta0 = log(mean(Y)/(1-mean(Y)))
	beta = matrix(0,p,1)	
	HHSVM.final = hhsvm(Z,2*Y-1,alpha=alpha, lambda=lambda.hhsvm_star,beta0, beta,niter=10000,maxVar=p)
	varSel.HHSVM.final = which(sign(HHSVM.final$beta) != 0)
	HHSVM.FP[repIndex] = sum(!varSel.HHSVM.final %in% 1:p.true)
	HHSVM.TP[repIndex] = sum(varSel.HHSVM.final %in% 1:p.true)
	HHSVM_CV_HX[[repIndex]]	= HHSVM.CV	
}

filename = paste("VariableSelectionExperiment_",nFolds,"fold.rda",sep="")
save(delta, seed, w, nc, alpha, x.sd, MLE.FP, MLE.TP, L2E.FP, L2E.TP, HHSVM.TP, HHSVM.FP,
	hhsvm.cv, glmnet.cv, rLogistic.cv, L2E.CV, MLE.CV, HHSVM.CV,
	L2E_models, MLE_models, HHSVM_models,
	L2E_CV_HX, MLE_CV_HX, HHSVM_CV_HX, 
	L2E_lambda_HX, MLE_lambda_HX, HHSVM_lambda_HX,
	file=filename)
