# Script for real data example in Section 8.1.
# "An n > p example: Predicting abnormal and normal vertebral columns."
# Data is from the UCI machine learning repository
# http://archive.ics.uci.edu/ml/datasets/Vertebral+Column

rm(list=ls())

Wdir = getwd()
Rdir = "../../Rfunctions"
setwd(Rdir)
source("mean.l2e.R")
source("stepSizeCalculator.R")
source("rLogistic.R")
source("calculateLambdaSequence.R")
setwd(Wdir)
library(glmnet)
library(ggplot2)
library(xtable)

vert3C = read.table("column_3C.dat", header=F, sep=" ")
ixDH = which(vert3C[,7] == "DH")
ixNO = which(vert3C[,7] == "NO")
ixSL = which(vert3C[,7] == "SL")

vert = read.table("column_2C.dat", header=F, sep=" ")
Y = ifelse(vert[,7] == "AB",1,0)
X = as.matrix(scale(vert[,-7]))
colnames(X) = c("PI","PT","LLA","SS","PR","GS")

### Scatter of Vertebrae data
Z = as.data.frame(X)
Z$Class = "DH"
Z$Class[ixNO] = "NO"
Z$Class[ixSL] = "SL"
Z$id = 1:nrow(Z)
Z$Label = factor(Y)
Z$Class = factor(Z$Class, levels=c("DH","SL","NO"),ordered=TRUE)
Z = melt(Z, id.vars=c("id","Label","Class"))

gre_colour = rgb(27/255, 158/255, 119/255)
brw_colour = rgb(216/255, 179/255, 101/255)

q = ggplot(Z, aes(x=id, y=value)) 
q = q + facet_wrap(~variable,ncol=3)
q = q + geom_point(aes(shape=Class, colour=Label))
q = q + scale_shape(name="Latent Class", breaks=c("DH","SL","NO"), labels=c("Disk Hernia","Spondylolisthesis","Normal"))
q = q + scale_shape_manual(value=c(3,2,1))
q = q + scale_colour_manual(name="Observed Class", values=c(brw_colour,gre_colour)
,breaks=c("1","0"),labels=c("Abnormal","Normal"))
q = q + ylab("Attribute Value") + xlab("Patient")
q + theme_bw() + opts(axis.title.x=theme_text(vjust=-0.25))

ggsave(filename="vertebrae_scatters.pdf",width=8, height=5)

#### Table of correlations between predictor variables.

xtable(cor(X), label="tab:vertebrae_cor",caption="Correlations among the six biomechanical attributes")

noi = 20000
w = 1; K = stepSizeCalculator(w)
p = ncol(X); n = nrow(X)

beta0 = log(mean(Y)/(1-mean(Y)))
beta = matrix(1,p,1)
alpha = 0.2

## Run the L2E fit forward

lambda = calculateLambdaSequence(Y,X,alpha=alpha,w=1,epsilon=0.00001,nLambda=100)

beta = matrix(0, p, 1)
beta0 = log(mean(Y)/(1-mean(Y)))

system.time({rLogistic.fit = rLogistic(X,Y,alpha,lambda,beta0,beta,maxVar=50, w=w)})

## Run the L2E fit backwards
lambda_first = rLogistic.fit$lambda[length(rLogistic.fit$lambda)]
lambda_last = rLogistic.fit$lambda[1]
lambda = 10^seq(log10(lambda_first), log10(lambda_last), length.out=100)

beta = matrix(1, p, 1)
beta0 = log(mean(Y)/(1-mean(Y)))
system.time({rLogistic.fit = rLogistic(X,Y,alpha,lambda,beta0,beta,niter=10000,maxVar=p, w=w)})

## Run GLMNET
glm.fit = glmnet(X,Y, family="binomial",alpha=alpha)

### Plot results with respect to J(beta)
J_cut_off = 5

penalties = apply(rLogistic.fit$beta, 2, FUN=function(x){alpha*sum(abs(x)) + 0.5*(1-alpha)*sum(x^2)})
rLogistic.df = cbind(penalties, t(rLogistic.fit$beta))
rLogistic.df = melt(as.data.frame(rLogistic.df),id.vars="penalties")
ix_truncated = which(rLogistic.df$penalties <= J_cut_off)
rLogistic.df = rLogistic.df[ix_truncated,]

penalties = apply(glm.fit$beta, 2, FUN=function(x){alpha*sum(abs(x)) + 0.5*(1-alpha)*sum(x^2)})
en.df = cbind(penalties,as.matrix(t(glm.fit$beta)))
en.df = melt(as.data.frame(en.df),id.vars="penalties")
ix_truncated = which(en.df$penalties <= J_cut_off)
en.df = en.df[ix_truncated,]
en.df$Method = "MLE"
rLogistic.df$Method = "L2E"
vertebrae.df = rbind(en.df, rLogistic.df)

q = ggplot(data=vertebrae.df, aes(x=penalties,y=value))
q = q + geom_line(aes(colour=variable, linetype=Method),size=1.25, alpha=0.8)
q = q + xlim(c(0,J_cut_off)) + ylab("Coefficients") + xlab(expression(paste("J(",beta,")")))
q = q + scale_colour_brewer(name="Attribute",palette="RdBu")
q = q + scale_linetype(breaks=c("MLE","L2E"),labels=c(expression(paste("MLE")),expression(paste(L[2],"E"))))
q + theme_bw() + opts(axis.title.x=theme_text(vjust=-0.25))
ggsave(filename="vertebrae_regularization_paths.pdf",width=8,height=3)
